/*
								+----------------------------------+
								|                                  |
								|  *** Basic compression algs ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2008   |
								|                                  |
								|            Compress.h            |
								|                                  |
								+----------------------------------+
*/

/*
 *	2007-02-25
 *
 *	this is first beta version if the file. todo - rewrite TBuffer so it can realloc
 *	itself in a way std::vector can and try to avoid any (re)allocations while (un)packing
 *
 *	2008-03-13
 *
 *	TBuffer was rewritten as requested, fixed some signed / unsigned mismatches for gcc
 *
 */

#ifndef __SIMPLE_COMPRESSION_INCLUDED
#define __SIMPLE_COMPRESSION_INCLUDED

#include "Integer.h"

/*
 *	struct TBuffer
 *		- simple data buffer structure (can hold up to 4GB on 32-bit machine)
 */
struct TBuffer {
protected:
	unsigned char *p_data;
	uint32_t n_size, n_capacity;

public:
	/*
	 *	TBuffer::TBuffer()
	 *		- default constructor, initializes empty buffer
	 */
	TBuffer();

	/*
	 *	TBuffer::TBuffer(uint32_t _n_size)
	 *		- creates a new buffer with size _n_size bytes
	 *		- note buffer is going to be empty in case there was not enough memory
	 */
	TBuffer(uint32_t _n_size);

	/*
	 *	TBuffer::TBuffer(const TBuffer &r_t_buffer)
	 *		- copy-constructor
	 *		- note buffer is going to be empty in case there was not enough memory
	 */
	TBuffer(const TBuffer &r_t_buffer);

	/*
	 *	TBuffer::~TBuffer()
	 *		- destructor
	 *		- deletes buffer data
	 */
	~TBuffer();

	/*
	 *	TBuffer &TBuffer::operator =(const TBuffer &r_t_buffer)
	 *		- copy-operator
	 *		- note buffer is going to be empty in case there was not enough memory
	 */
	TBuffer &operator =(const TBuffer &r_t_buffer);

	/*
	 *	void TBuffer::Swap(TBuffer &r_t_buffer)
	 *		- swaps contents of this and r_t_buffer
	 */
	void Swap(TBuffer &r_t_buffer);

	/*
	 *	bool TBuffer::Resize(uint32_t n_new_size, bool b_data_important = true)
	 *		- changes size of the buffer
	 *		- note that resizing below buffer capacity never causes reallocations
	 *		- if b_data_important is set, buffer data are copied when reallocating,
	 *		  otherwise they're not (speed optimization)
	 *		- returns true on success, false on failure (data aren't lost on failure,
	 *		  but buffer size remains unchaged)
	 */
	bool Resize(uint32_t n_new_size, bool b_data_important = true);

	/*
	 *	bool TBuffer::Grow(uint32_t n_add)
	 *		- resizes the buffer so it contains at least n_add more free bytes
	 *		- returns true on success, false on failure (data aren't lost on failure,
	 *		  but buffer size remains unchaged)
	 */
	bool Grow(uint32_t n_add);

	/*
	 *	void TBuffer::Clear()
	 *		- clears contents of the buffer (both size and capacity will be 0)
	 */
	void Clear();

	/*
	 *	inline bool b_Empty() const
	 *		- returns true in case buffer is empty (size is 0), otherwise false
	 */
	inline bool b_Empty() const
	{
		return !n_size;
	}

	/*
	 *	inline uint32_t n_Size() const
	 *		- returns size of the buffer
	 */
	inline uint32_t n_Size() const
	{
		return n_size;
	}

	/*
	 *	inline uint32_t n_Capacity() const
	 *		- returns capacity of the buffer
	 */
	inline uint32_t n_Capacity() const
	{
		return n_capacity;
	}

	/*
	 *	inline unsigned char *p_Data()
	 *		- returns data buffer
	 */
	inline unsigned char *p_Data()
	{
		return p_data;
	}

	/*
	 *	inline const unsigned char *p_Data() const
	 *		- returns data buffer for reading
	 */
	inline const unsigned char *p_Data() const
	{
		return p_data;
	}

	/*
	 *	uint32_t TBuffer::n_Max_Size()
	 *		- returns the greatest size possible buffer can be allocated to
	 */
	static uint32_t n_Max_Size();
};

/*
 *	CBurrowsWheelerTransform
 *		- simple Burrows-Wheeler transform implementation
 *		- uses indices instead of string copies to minimize memory usage
 */
class CBurrowsWheelerTransform {
protected:
	class CWrapMemCmp;
	class CIota;
	class CSorter;
	class CMerger;

public:
	/*
	 *	static bool CBurrowsWheelerTransform::Decode(const TBuffer &r_t_in_buffer,
	 *		TBuffer &r_t_out_buffer)
	 *		- decodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
	 *		- returns true on success, false on failure
	 */
	static bool Decode(const TBuffer &r_t_src, TBuffer &r_t_dest);

	/*
	 *	static bool CBurrowsWheelerTransform::Encode(const TBuffer &r_t_in_buffer,
	 *		TBuffer &r_t_out_buffer)
	 *		- encodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
	 *		- returns true on success, false on failure
	 *		- note this doesn't work with empty input buffer
	 */
	static bool Encode(const TBuffer &r_t_src, TBuffer &r_t_dest);

	/*
	 *	static bool CBurrowsWheelerTransform::ThreadedEncode(const TBuffer &r_t_in_buffer,
	 *		TBuffer &r_t_out_buffer, int n_thread_num)
	 *		- encodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
	 *		- works in parallel, n_thread_num must be power of two
	 *		- returns true on success, false on failure
	 *		- note this doesn't work with empty input buffer
	 */
	static bool ThreadedEncode(const TBuffer &r_t_src, TBuffer &r_t_dest, int n_thread_num);
};

/*
 *	CMoveToFrontTransform
 *		- basic move to front transformation implementation
 */
class CMoveToFrontTransform {
public:
	/*
	 *	CMoveToFrontTransform::<unnamed>
	 *		- MTF algorithm names
	 */
	enum {
		algo_MTF,
		algo_MTF_1
	};

	/*
	 *	static void CMoveToFrontTransform::Decode(const TBuffer &r_t_buffer,
	 *		int n_algorithm = algo_MTF)
	 *		- decodes data in r_t_buffer (operates in-sit)
	 *		- n_algorithm is MTF algorithm, one of algo_MTF, algo_MTF_1
	 */
	static void Decode(TBuffer &r_t_buffer, int n_algorithm = algo_MTF);

	/*
	 *	static void CMoveToFrontTransform::Encode(const TBuffer &r_t_buffer,
	 *		int n_algorithm = algo_MTF)
	 *		- encodes data in r_t_buffer (operates in-sit)
	 *		- n_algorithm is MTF algorithm, one of algo_MTF, algo_MTF_1
	 */
	static void Encode(TBuffer &r_t_buffer, int n_algorithm = algo_MTF);

	/*
	 *	static bool CMoveToFrontTransform::Decode(const TBuffer &r_t_in_buffer,
	 *		TBuffer &r_t_out_buffer, int n_algorithm = algo_MTF)
	 *		- decodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
	 *		- n_algorithm is MTF algorithm, one of algo_MTF, algo_MTF_1
	 *		- returns true on success, false on failure
	 */
	static bool Decode(const TBuffer &r_t_src, TBuffer &r_t_dest, int n_algorithm = algo_MTF);

	/*
	 *	static bool CMoveToFrontTransform::Encode(const TBuffer &r_t_in_buffer,
	 *		TBuffer &r_t_out_buffer, int n_algorithm = algo_MTF)
	 *		- encodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
	 *		- n_algorithm is MTF algorithm, one of algo_MTF, algo_MTF_1
	 *		- returns true on success, false on failure
	 */
	static bool Encode(const TBuffer &r_t_src, TBuffer &r_t_dest, int n_algorithm = algo_MTF);

protected:
	static void _Encode(const TBuffer &r_t_src, TBuffer &r_t_dest, int n_algorithm);
	static void _Decode(const TBuffer &r_t_src, TBuffer &r_t_dest, int n_algorithm);
};

/*
 *	class CRunLengthCodec
 *		- simple run length coder, optimized for packing MTF outputs
 *		  (compression flag bit is LSB so, in theory, symbols with lower values are
 *		  generated, in hope not to disrupt symbol probabilities after MTF)
 */
class CRunLengthCodec {
protected:
	enum {
		min_RunLength = 2,	// minimal length 
		rle_Zero = false	// use RLE-0? (only encode runs of nulls)
	};

public:
	/*
	 *	static bool CRunLengthCodec::Decode(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer)
	 *		- decodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
	 *		- returns true on success, false on failure
	 */
	static bool Decode(const TBuffer &r_t_src, TBuffer &r_t_dest);

	/*
	 *	static bool CRunLengthCodec::Encode(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer)
	 *		- encodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
	 *		- returns true on success, false on failure
	 */
	static bool Encode(const TBuffer &r_t_src, TBuffer &r_t_dest);
};

/*
 *	class CHuffmanCodec
 *		- very simple and easy to use cannonical huffman (de)coder, operating on bytes
 *		- stores (uncompressed) file length, symbol counts and symbols,
 *		  followed by bitstream. last byte is padded by zeros.
 */
class CHuffmanCodec {
protected:
	enum {
		max_CodeBitNum = 16	// can be 1 to 31
	};

	struct TFrequency;
	struct TNode;

public:
	/*
	 *	static bool CHuffmanCodec::Decode(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer)
	 *		- decodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
	 *		- returns true on success, false on failure
	 */
	static bool Decode(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer);

	/*
	 *	static bool CHuffmanCodec::Encode(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer)
	 *		- encodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
	 *		- returns true on success, false on failure
	 */
	static bool Encode(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer);

protected:
	static inline bool CompareSymbol(const TFrequency &r_freq_a, const TFrequency &r_freq_b);
	static inline bool FindZeroFreq(const TFrequency &r_t_freq);
	static bool CreateHuffmanTree(std::vector<TFrequency> &r_freq_list, int *p_code_num);
};

#endif //__SIMPLE_COMPRESSION_INCLUDED
